package utils

import (
	"bytes"
	"context"
	"encoding/binary"
	"errors"
	"fmt"
	"git.gammaspectra.live/P2Pool/consensus/v3/merge_mining"
	"git.gammaspectra.live/P2Pool/consensus/v3/monero/block"
	"git.gammaspectra.live/P2Pool/consensus/v3/monero/client"
	"git.gammaspectra.live/P2Pool/consensus/v3/monero/client/rpc/daemon"
	"git.gammaspectra.live/P2Pool/consensus/v3/monero/crypto"
	"git.gammaspectra.live/P2Pool/consensus/v3/monero/transaction"
	"git.gammaspectra.live/P2Pool/consensus/v3/p2pool/sidechain"
	"git.gammaspectra.live/P2Pool/consensus/v3/types"
	"git.gammaspectra.live/P2Pool/consensus/v3/utils"
	"git.gammaspectra.live/P2Pool/observer-cmd-utils/index"
	"github.com/hashicorp/golang-lru/v2"
	"math"
)

// keep last few main blocks
var mainBlockCache, _ = lru.New[types.Hash, *block.Block](100)

func GetMainBlock(id types.Hash, client *client.Client) (*block.Block, error) {
	if bb, ok := mainBlockCache.Get(id); !ok {
		blockData, err := client.GetBlock(id, context.Background())
		if err != nil {
			return nil, err
		}
		mainBlock := &block.Block{}
		err = mainBlock.UnmarshalBinary(blockData.Blob, false, nil)
		if err != nil {
			return nil, err
		}
		if len(mainBlock.Coinbase.Outputs) == 0 {
			//should never happen from monero itself
			return nil, errors.New("outputs not filled")
		}
		mainBlockCache.Add(id, mainBlock)
		return mainBlock, nil
	} else {
		return bb, nil
	}
}

func FindAndInsertMainHeader(h daemon.BlockHeader, indexDb *index.Index, storeFunc func(b *sidechain.PoolBlock), client *client.Client, difficultyByHeight block.GetDifficultyByHeightFunc, getByTemplateId sidechain.GetByTemplateIdFunc, getByMainId sidechain.GetByMainIdFunc, getByMainHeight sidechain.GetByMainHeightFunc, processBlock func(b *sidechain.PoolBlock) error) error {
	var indexMain *index.MainBlock
	indexMain = indexDb.GetMainBlockByHeight(h.Height)
	if indexMain == nil {
		indexMain = &index.MainBlock{}
	}
	if indexMain.Id != h.Hash {
		//Zero data if they don't match to insert new height
		indexMain.Metadata = make(map[string]any)
		indexMain.SideTemplateId = types.ZeroHash
		indexMain.CoinbasePrivateKey = crypto.PrivateKeyBytes{}
	}
	indexMain.Id = h.Hash
	indexMain.Height = h.Height
	indexMain.Timestamp = uint64(h.Timestamp)
	indexMain.Reward = h.Reward
	indexMain.CoinbaseId = h.MinerTxHash
	//todo: top64 bits
	indexMain.Difficulty = h.Difficulty

	// Find possible candidates that match on main height, main previous id, header nonce, timestamp, and count of transactions to reduce full block lookups
	var candidates sidechain.UniquePoolBlockSlice
	for _, b := range getByMainHeight(h.Height) {
		if b.Main.PreviousId == h.PrevHash && b.Main.Coinbase.AuxiliaryData.TotalReward == h.Reward && b.Main.Timestamp == uint64(h.Timestamp) && len(b.Main.Transactions) == int(h.NumTxes) {
			// candidate for checking
			candidates = append(candidates, b)
		}
	}

	if len(candidates) == 0 {
		// no candidates, insert header as-is
		if err := indexDb.InsertOrUpdateMainBlock(indexMain); err != nil {
			return err
		}
		return nil
	}

	utils.Logf("Monero", "checking block %s at height %d: %d candidate(s)", h.Hash, h.Height, len(candidates))

	mainBlock, err := GetMainBlock(h.Hash, client)
	if err != nil {
		utils.Errorf("Monero", "could not get main block: %e", err)
		// insert errored block as-is
		if err := indexDb.InsertOrUpdateMainBlock(indexMain); err != nil {
			return err
		}
		return nil
	}

	var possibleId types.Hash

	var extraNonce []byte

	// try to fetch extra nonce if existing
	{
		if t := mainBlock.Coinbase.Extra.GetTag(uint8(sidechain.SideExtraNonce)); t != nil {
			if len(t.Data) >= sidechain.SideExtraNonceSize || len(t.Data) < sidechain.SideExtraNonceMaxSize {
				extraNonce = t.Data
			}
		}
	}

	// get a merge tag if existing
	{
		mergeMineTag := mainBlock.Coinbase.Extra.GetTag(transaction.TxExtraTagMergeMining)
		if mergeMineTag != nil {
			func() {
				shareVersion := sidechain.P2PoolShareVersion(indexDb.Consensus(), mainBlock.Timestamp)
				if shareVersion < sidechain.ShareVersion_V3 {
					//TODO: this is to comply with non-standard p2pool serialization, see https://github.com/SChernykh/p2pool/issues/249
					if mergeMineTag.VarInt != types.HashSize {
						return
					}

					if len(mergeMineTag.Data) != types.HashSize {
						return
					}

					// sidechain id
					possibleId = types.HashFromBytes(mergeMineTag.Data)
				} else {
					// fallback for "old" non forked blocks
					if len(mergeMineTag.Data) == types.HashSize && mergeMineTag.VarInt == types.HashSize {
						// sidechain id
						possibleId = types.HashFromBytes(mergeMineTag.Data)
						return
					}

					//properly decode merge mining tag
					mergeMineReader := bytes.NewReader(mergeMineTag.Data)
					var mergeMiningTag merge_mining.Tag
					if err := mergeMiningTag.FromReader(mergeMineReader); err != nil {
						return
					}
					if mergeMineReader.Len() != 0 {
						return
					}

					possibleId = mergeMiningTag.RootHash
				}
			}()
		}

	}

	if possibleId == types.ZeroHash || len(extraNonce) == 0 {
		utils.Logf("Monero", "checking block %s at height %d: not a p2pool block", h.Hash, h.Height)
		if err := indexDb.InsertOrUpdateMainBlock(indexMain); err != nil {
			return err
		}
		return nil
	}

	if t := getByTemplateId(possibleId); t != nil {
		//found template
		if err := processBlock(t); err != nil {
			return fmt.Errorf("could not process template: %w", err)
		}

		if t.MainId() == h.Hash { //exact match, no need to re-arrange template inclusion
			sideBlock := indexDb.GetSideBlockByMainId(t.MainId())
			if sideBlock == nil {
				//found deep orphan which found block, insert parents
				cur := t
				for cur != nil && !index.QueryHasResults(indexDb.GetSideBlocksByTemplateId(cur.SideTemplateId(indexDb.Consensus()))) {
					orphanBlock, orphanUncles, err := indexDb.GetSideBlockFromPoolBlock(cur, index.InclusionOrphan)
					if err != nil {
						break
					}

					if index.QueryHasResults(indexDb.GetSideBlocksByTemplateId(orphanBlock.TemplateId)) {
						continue
					}
					if err := indexDb.InsertOrUpdateSideBlock(orphanBlock); err != nil {
						return fmt.Errorf("error inserting %s, %s at %d: %s", cur.SideTemplateId(indexDb.Consensus()), cur.MainId(), cur.Side.Height, err)
					}

					for _, orphanUncle := range orphanUncles {
						if index.QueryHasResults(indexDb.GetSideBlocksByTemplateId(orphanUncle.TemplateId)) {
							continue
						}
						if err := indexDb.InsertOrUpdateSideBlock(orphanUncle); err != nil {
							return fmt.Errorf("error inserting uncle of %s, %s; %s, %s at %d: %s", cur.SideTemplateId(indexDb.Consensus()), cur.MainId(), orphanUncle.TemplateId, orphanUncle.MainId, cur.Side.Height, err)
						}
					}

					cur = getByTemplateId(cur.Side.Parent)
					if cur == nil {
						break
					}
					if err := processBlock(cur); err != nil {
						return fmt.Errorf("could not process cur template: %w", err)
					}
				}
				sideBlock = indexDb.GetSideBlockByMainId(t.MainId())
				if sideBlock == nil {
					return errors.New("no block in database for orphan")
				}
				sideBlock.Inclusion = index.InclusionOrphan
			} else if sideBlock.Inclusion == index.InclusionAlternateInVerifiedChain {
				sideBlock.Inclusion = index.InclusionInVerifiedChain
			}

			if err := indexDb.InsertOrUpdateSideBlock(sideBlock); err != nil {
				return fmt.Errorf("error inserting %s, %s at %d: %s", sideBlock.TemplateId, sideBlock.MainId, sideBlock.SideHeight, err)
			}
			t.Depth.Store(math.MaxUint64)
			storeFunc(t)

			indexMain.SideTemplateId = t.FastSideTemplateId(indexDb.Consensus())
			indexMain.RootHash = t.MergeMiningTag().RootHash

			indexMain.CoinbasePrivateKey = t.Side.CoinbasePrivateKey
			if err := indexDb.InsertOrUpdateMainBlock(indexMain); err != nil {
				return err
			}
			utils.Logf("Monero", "INSERTED found block %s at height %d, template id %s", h.Hash, h.Height, indexMain.SideTemplateId)
			if err := FindAndInsertMainHeaderOutputs(indexMain, indexDb, client, difficultyByHeight, getByTemplateId, getByMainId, getByMainHeight, processBlock); err != nil {
				utils.Errorf("", "error inserting coinbase outputs: %s", err)
			}
			return nil
		} else {
			data, _ := t.MarshalBinary()
			newBlock := &sidechain.PoolBlock{}
			if err := newBlock.UnmarshalBinary(indexDb.Consensus(), &sidechain.NilDerivationCache{}, data); err != nil {
				utils.Panic(err)
			}
			copy(newBlock.CoinbaseExtra(sidechain.SideExtraNonce), extraNonce)
			newBlock.Main.Nonce = uint32(h.Nonce)
			newBlock.Depth.Store(math.MaxUint64)

			if newBlock.MainId() == h.Hash {
				//store into archive
				storeFunc(newBlock)
				//insert into db as well
				var sideBlock *index.SideBlock
				if indexDb.GetSideBlockByMainId(t.MainId()).Inclusion == index.InclusionOrphan {
					sideBlock, _, err = indexDb.GetSideBlockFromPoolBlock(newBlock, index.InclusionOrphan)
				} else {
					sideBlock, _, err = indexDb.GetSideBlockFromPoolBlock(newBlock, index.InclusionInVerifiedChain)
				}
				if err != nil {
					return fmt.Errorf("could not process %s, at side height %d, template id %s: %w", t.FastSideTemplateId(indexDb.Consensus()), t.Side.Height, t.FastSideTemplateId(indexDb.Consensus()), err)
				}

				if err := indexDb.InsertOrUpdateSideBlock(sideBlock); err != nil {
					return fmt.Errorf("error inserting %s, %s at %d: %s", sideBlock.TemplateId, sideBlock.MainId, sideBlock.SideHeight, err)
				}

				indexMain.SideTemplateId = t.FastSideTemplateId(indexDb.Consensus())
				indexMain.RootHash = t.MergeMiningTag().RootHash

				indexMain.CoinbasePrivateKey = newBlock.Side.CoinbasePrivateKey
				if err := indexDb.InsertOrUpdateMainBlock(indexMain); err != nil {
					return err
				}
				utils.Logf("", "INSERTED ALTERNATE found block %s at height %d, template id %s", h.Hash, h.Height, indexMain.SideTemplateId)
				if err := FindAndInsertMainHeaderOutputs(indexMain, indexDb, client, difficultyByHeight, getByTemplateId, getByMainId, getByMainHeight, processBlock); err != nil {
					utils.Errorf("", "error inserting coinbase outputs: %s", err)
				}
				return nil
			}
		}
	}

	utils.Logf("", "checking block %s at height %d, possible id %s: could not find matching possible id", h.Hash, h.Height, possibleId)

	// could not find template, maybe it's other pool?

	// fill template id for future reference
	// TODO: do this for all blocks
	indexMain.SetMetadata("merge_mining_tag", possibleId)
	indexMain.SetMetadata("extra_nonce", binary.LittleEndian.Uint32(extraNonce))

	if err := indexDb.InsertOrUpdateMainBlock(indexMain); err != nil {
		return err
	}
	return nil
}

func FindAndInsertMainHeaderOutputs(mb *index.MainBlock, indexDb *index.Index, client *client.Client, difficultyByHeight block.GetDifficultyByHeightFunc, getByTemplateId sidechain.GetByTemplateIdFunc, getByMainId sidechain.GetByMainIdFunc, getByMainHeight sidechain.GetByMainHeightFunc, processBlock func(b *sidechain.PoolBlock) error) error {
	if !index.QueryHasResults(indexDb.GetMainCoinbaseOutputs(mb.CoinbaseId)) {
		//fill information
		utils.Logf("", "inserting coinbase outputs for %s, template id %s, coinbase id %s", mb.Id, mb.SideTemplateId, mb.CoinbaseId)
		var t *sidechain.PoolBlock
		if t = getByTemplateId(mb.SideTemplateId); t == nil || t.MainId() != mb.Id {
			t = nil
		}
		if t == nil {
			t = getByMainId(mb.Id)
		}
		if t != nil {
			if err := processBlock(t); err != nil {
				return fmt.Errorf("could not process block: %s", err)
			}
			if mb.CoinbaseId != t.CoinbaseId() {
				return fmt.Errorf("not matching coinbase id: %s vs %s", mb.CoinbaseId, t.CoinbaseId())
			}
			indexes, err := client.GetOutputIndexes(mb.CoinbaseId)
			if err != nil {
				return fmt.Errorf("error getting output indexes: %w", err)
			}
			if len(indexes) != len(t.Main.Coinbase.Outputs) {
				return fmt.Errorf("not matching indexes vs coinbase output len: %d vs %d", len(indexes), len(t.Main.Coinbase.Outputs))
			}
			preAllocatedShares := sidechain.PreAllocateShares(indexDb.Consensus().ChainWindowSize * 2)
			shares, _ := sidechain.GetShares(t, indexDb.Consensus(), difficultyByHeight, getByTemplateId, preAllocatedShares)
			if len(shares) != len(t.Main.Coinbase.Outputs) {
				return fmt.Errorf("not matching shares vs coinbase output len: %d vs %d", len(shares), len(t.Main.Coinbase.Outputs))
			}
			outputs := make(index.MainCoinbaseOutputs, 0, len(indexes))
			for _, o := range t.Main.Coinbase.Outputs {
				output := index.MainCoinbaseOutput{
					Id:                mb.CoinbaseId,
					Index:             uint32(o.Index),
					GlobalOutputIndex: indexes[o.Index],
					Miner:             indexDb.GetOrCreateMinerPackedAddress(shares[o.Index].Address).Id(),
					Value:             o.Reward,
				}
				outputs = append(outputs, output)
			}
			if err := indexDb.InsertOrUpdateMainCoinbaseOutputs(outputs); err != nil {
				return err
			}
		} else {
			return errors.New("nil template")
		}
	}
	return nil
}
